/*
   Copyright (C) 2006 by Alexei Krasnopolski

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License. 
*/
package com.ak.osgi.service.rmi.entrant;

import java.io.IOException;
import java.lang.reflect.Method;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Timer;
import java.util.TimerTask;

import org.knopflerfish.service.log.LogRef;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.util.tracker.ServiceTrackerCustomizer;

import com.ak.osgi.service.rmi.RmiConstants;
import com.ak.osgi.service.rmi.interfaces.rmifwcontext.RemoteOSGiFramework;
import com.ak.osgi.service.rmi.remotefw.BundleShortInfoBean;
import com.ak.osgi.service.rmi.remotefw.RMIEntrant;
import com.ak.osgi.service.rmi.remotefw.RMIGate;
import com.ak.osgi.service.rmi.remotefw.RMIServiceBean;
import com.ak.osgi.service.rmi.utility.DebugLog;

/**
 * The <tt>Client</tt> class implements <tt>RMIEntrant</tt>. After adding <tt>Client</tt> 
 * to RMIGate object on remote framework one listens
 * to remote framework. Every service registered on remote framework with 
 * non empty "rmiservice" property will be registered by <tt>Client</tt> on this local framework  
 *  
 * @version 1.0.0
 * @author <a href="mailto:krasnop@bellsouth.net">Alexei Krasnopolski</a>
 */
public class Client extends UnicastRemoteObject implements RMIEntrant
{
  private static final long serialVersionUID = 1L;
  private static int currentIdx = 0;
  private int idx;
  /**
   *  This field is a reference to <tt>BundleActivator</tt> object that was running this object.
   */
//  com.ak.osgi.service.rmi.Activator activator;
  /**
   *  <tt>BundleContext</tt> of the <tt>Client</tt> bundle.
   */
  BundleContext bc;
  /**
   *  This field is remote reference to <tt>RMIGate</tt> object on remote framework 
   *  connected to the <tt>Client</tt>. 
   */
  RMIGate gate;
  /**
   *  Name of remote framework host.  
   */
  private String host;
  /**
   *  Port of RMI registry running on remote host. The <tt>RMIGate</tt> object of 
   *  remote framework have to be registered on the RMI registry
   */
  private int port;
  /**
   *  A <tt>ServiceTracker</tt> object is created to listen the changes of remote services.
   */
  private ServiceTracker rmiServiceTracker;
  /**
   *  Table of a references of <tt>ServiceRegistration</tt> objects for remote services 
   *  registered on this local framework.
   *  <br>key - hash value of <tt>Integer</tt> type;
   *  <br>value - <tt>ServiceRegistration</tt> object after registration of remote service 
   *  on local framework. 
   */
  private Hashtable serviceReg;
  /**
   *  The field helps to control of the object after STOPPING event   
   */
  boolean isStoppingMode = false;
  boolean connected = false;
  Timer timer;
  TimerTask task;
  LogRef log;
  /**
   * The <tt>RMIServiceTrackerCustomizer</tt> class provides processing 
   * of service life cycle events.
   * <p>If instance of <tt>RMIGate</tt> service is adding 
   * that <tt>gate</tt> field assigns reference to this service.
   * <p>If service of <tt>RMIGate</tt> type is removing from the local framework the bundle is stopped. 
   * 
   */
  class RMIServiceTrackerCustomizer implements ServiceTrackerCustomizer
  {
    public Object addingService( ServiceReference reference )
    {
      Object so = bc.getService( reference );
      String errmsg = "";
      if( so instanceof RMIGate && gate == null )
      {
        try
        {
          errmsg = "CLIENT > Entrant exception: can not cast to (RMIGate)";
          gate = (RMIGate)so;
        }
        catch( ClassCastException cce )
        {
          log.error( errmsg, reference, cce );
        }
      }
      if( so instanceof RemoteOSGiFramework )
      {
        connected = true;
      }
      String index = (String)reference.getProperty( RmiConstants.KEY_RMISERVICE_CLIENT_IDX );
      log.info("CLIENT > added RMI service : " + so.getClass().getName() 
          + " hash #" + reference.hashCode()
          + " idx=" + index );
      DebugLog.logRemoteInstance( log, "service(remote)", (Remote)so );
      return so;
    }

    public void modifiedService( ServiceReference reference, Object service )
    {}

    public void removedService( ServiceReference reference, Object service )
    {        
      bc.ungetService( reference );
      if( service instanceof RMIGate && (!isStoppingMode) )
      {
        freeze();
      }
      if( service instanceof RemoteOSGiFramework )
      {
        connected = false;
      }
      log.info("CLIENT > removed RMI service : " + service.getClass().getName() + " hash #" + reference.hashCode() );
    }  
  }

  /**
   * Create instance of <tt>Client</tt> object
   * @param $host 
   *           Name of remote fromework host. 
   * @param $port
   *           Port of rmiregistry running on remote host.
   * @param $bc          
   *           Bundle context 
   * @throws RemoteException
   */
  public Client( String $host, int $port, com.ak.osgi.service.rmi.Activator $activator ) throws RemoteException
  {
    this( $host, $port, $activator.getBundleContext(), $activator.getLogger() );
//    bc = $activator.getBundleContext();
//    log = $activator.getLogger();
    
//    gate = null;
//    host = $host;
//    port = $port;
//    serviceReg = new Hashtable();
//    idx = currentIdx++;
  }

  public Client( String $host, int $port, BundleContext $bc, LogRef $log ) throws RemoteException
  {
    super();
    bc = $bc;
    log = $log;

    gate = null;
    host = $host;
    port = $port;
    serviceReg = new Hashtable();
    idx = currentIdx++;
    DebugLog.logRemoteInstance( log, "client(local)", this );
//    ClassLoader cl = this.getClass().getClassLoader();
//    log.debug( "client stub = " + this );
//    log.debug( "class loader = " + cl );
//    log.debug( "parent class loader = " + cl.getParent() ); 
//    this.setLog( System.out );
  }

  /**
   * Create instance of default <tt>Client</tt> object
   * @param $bc          
   *           Bundle context 
   * @throws RemoteException
   */
  public Client( com.ak.osgi.service.rmi.Activator $activator ) throws RemoteException
  {
    this( "localhost", 1199, $activator );
  }

  public int getIdx()
  {
    return idx;  
  }
  
  public boolean isConnected()
  {
    return connected;  
  }
  
  public void startOpening()
  {
    timer = new Timer();
    task = new TimerTask()
    {
      public void run()
      {
        try
        {
          if( Client.this.getGate() == null )
          {
            Client.this.open();
            log.info( "CLIENT > RMI Client start." );
          }
          cancel();
        }
        catch( Exception ioe )
        {
          log.debug( "CLIENT > Can not connect to remote Gate as a client." );
        }
        
      }
    };
    timer.schedule( task, RmiConstants.DELAY, RmiConstants.PERIOD );
  }
  
  /**
   * Open the <tt>Client</tt> object:
   *  <br>1. Connect to rmi registry on remote host and obtain remote reference to <tt>RMIGate</tt>,
   *  <br>2. Add itself as a client to remote RMI Gate,
   *  <br>3. Create and open service tracker for remote service <tt>RMIGate</tt>,
   *  <br>4. After that reobtain instance of <tt>RMIGate</tt> as remote service now.
   * @throws Exception
   *           1. If <tt>Client</tt> can not connect to RMI Gate
   *           2. If <tt>ServiceTracker</tt> object can not be created or opened. 
   *           3. If <tt>RMIGate</tt> object is null
   */
  public void open() throws Exception
  {
    if( connect() == null )
      throw new IOException( "CLIENT > Can not connect to RMI Gate." );
    String errmsg = "";
    try
    {
      errmsg = "CLIENT > can not add RMIEntrant to RMIGate: ";
      gate.addClient( this );

//      String rmiServiceFilter = "(|(&(" + Constants.OBJECTCLASS + "=com.ak.osgi.service.rmi.remotefw.RMIGate)(" 
//        + RmiConstants.KEY_RMISERVICE_REM + "=*))";
//      rmiServiceFilter += "(&(" + Constants.OBJECTCLASS + "=com.ak.osgi.service.rmi.interfaces.rmifwcontext.RemoteOSGiFramework)(" 
//      + RmiConstants.KEY_RMISERVICE_REM + "=*)))";
      String rmiServiceFilter = "("
        +"&(" 
        + "|(" + Constants.OBJECTCLASS + "=" + RMIGate.class.getName() + ")" 
        + "(" + Constants.OBJECTCLASS + "=" + RemoteOSGiFramework.class.getName() + ")" 
        + ")"
        + "(" + RmiConstants.KEY_RMISERVICE_REM + "=*)" 
        + "(" + RmiConstants.KEY_RMISERVICE_CLIENT_IDX + "=" + String.valueOf( idx ) + ")" 
        + ")";

      errmsg = "CLIENT > Failed to create filter: " + rmiServiceFilter;
      rmiServiceTracker = new ServiceTracker( bc, bc.createFilter( rmiServiceFilter ), new RMIServiceTrackerCustomizer() );
      errmsg = "CLIENT > Failed to set up rmiService tracker: ";
      gate = null;
      rmiServiceTracker.open();
//      errmsg = "CLIENT > can not get RMIGate service from service tracker: ";
//      gate = (RMIGate)rmiServiceTracker.getService();
//      if( gate == null )
//        throw new Exception( "CLIENT > gate = null" );
//      debugDump( gate );
    }
    catch( Exception e )
    {
      log.error( "CLIENT > Entrant exception: " + errmsg, e );
      throw e;
    }
    isStoppingMode = false;
  }

  /**
   * Obtain reference to rmi registry on remote host and look up for remote object by name 'RMIGate'
   * 
   * @return
   *      reference to rmi registry on remote host
   * @throws NoSuchMethodException 
   * @throws SecurityException 
   */
  private Registry connect()
  {
    String errmsg = "";
    Registry registry = null;
    String [] list = new String[0];
    
    try
    {
      errmsg = "can not get registry: " + host + ":" + port;
      registry = LocateRegistry.getRegistry( host, port );
      list = registry.list();
      errmsg = "can not find 'RMIGate' on rmi-registry (" + host + ":" + port + ")";
      gate = (RMIGate)registry.lookup( "RMIGate" );
      log.info( "CLIENT > the client has connected to Remote Framework (RMI Gate)." );
      DebugLog.logRemoteInstance( log, "gate(remote)", gate );
    }
    catch( NotBoundException e )
    {
      log.warn( "CLIENT > Entrant can not lookup 'RMI Gate'.", e );
    }
    catch( RemoteException e )
    {
      log.warn( "CLIENT > Entrant " + errmsg, e );
      for( int i = 0; i < list.length; i++ )
        System.out.println( "(" + i + ") " + list[i] );
    }

    if( gate == null )
      return null;
    return registry;
  }
  
  /**
   * Reset connection to RMI Gate
   */
  void disconnect()
  {
    serviceReg.clear();
    task.cancel();
    timer.cancel();
    gate = null;
  }
  
  /**
   * Close the <tt>Client</tt> object:
   *  <p>1. Close service tracker for remote service <tt>RMIGate</tt>,
   *  <p>2. Remove itself as a client from remote RMI Gate,
   *  <p>3. disconnect from <tt>RMIGate</tt> remote object,
   * @throws Exception
   */
  public void close()
  {
    isStoppingMode = true;
    String errmsg = "stop with rmiServiceTracker reference = null.";
    if( rmiServiceTracker != null )
    {
      rmiServiceTracker.close(); 
    }
    else
      log.warn( "CLIENT > Entrant message: " + errmsg );
    try
    {
      errmsg = "stop with object RMI Gate = null.";
      if( gate != null )
      {  
        errmsg = "while remove this entrant from RMI Gate.";
        gate.removeClient( this );
      }
      else
        log.warn( "CLIENT > Entrant message: " + errmsg );
    }
    catch( Exception e )
    {
      log.error( "CLIENT > Entrant exception: " + errmsg, e );
    }
    disconnect();
  }
  
  protected void freeze()
  {
    String errmsg = "";
    isStoppingMode = false;
    try
    {
      errmsg = "freeze with rmiServiceTracker reference = null.";
      rmiServiceTracker.close();
    }
    catch( Exception e )
    {
      log.error( "CLIENT > Entrant exception: " + errmsg, e );
    }
    gate = null;//disconnect();
    startOpening();
  }

  /**
   * The method implements <tt>RMIEntrant</tt> interface.
   * RMIGate object invokes this method to unregister remote service of remote framework from local framework.
   * 
   *@param hashCode
   *          Key of remote service in <tt>serviceReg</tt> hashtable .
   *@throws RemoteException
   *          If any exceptions from <tt>unregister()</tt> occures.
   *@see com.ak.osgi.service.rmi.remotefw.RMIEntrant#unregisterRMIService(int)
   */ 
  public void unregisterRMIService( int hashCode ) throws RemoteException
  {
    ServiceRegistration sreg = (ServiceRegistration)serviceReg.get( new Integer( hashCode ) );
    if( sreg != null )
    {
      sreg.unregister();     // after this call the bundle can come to stopping mode 
      if( ! isStoppingMode )
        log.info( "CLIENT > service is unregistered : hash #" + hashCode );
    }
    else
      throw new RemoteException( "CLIENT > Entrant's service registration = null..." );
  }

  public RMIGate getGate()
  {
    return gate;
  }

  /**
   * The method implements <tt>RMIEntrant</tt> interface.
   * RMIGate object invokes this method to register remote service of remote framework to local framework.
   * 
   *@param bean
   *          java bean instance of <tt>RMIServiceBean</tt> that contents:<br>
   *        hash Code - unique id (on server FW) of the remote service;<br> 
   *        clazzes -  List of interfaces for <tt>rmiService</tt>;<br>
   *        rmiService - Remote service to register on client framework;<br>
   *        properties - Property hashtable for the remote service.
   *@throws RemoteException
   *@see com.ak.osgi.service.rmi.remotefw.RMIEntrant#registerRMIService(com.ak.osgi.service.rmi.remotefw.RMIServiceBean)
   */ 
  
  public void registerRMIService( RMIServiceBean bean ) throws RemoteException
  {
    ServiceRegistration sreg = null;
    Remote so = null;
    try
    {
      Dictionary d = bean.getProperties();
      d.put( RmiConstants.KEY_RMISERVICE_CLIENT_IDX, String.valueOf( idx ) );
      so = bean.getRMIService();
      sreg = bc.registerService( bean.getClasses(), so, d );
      serviceReg.put( new Integer( bean.getHashCode() ), sreg );
    }
    catch( Exception e )
    {
      log.error( "CLIENT > Entrant exception: while register remote service " + bean.getClasses()[0], e );
      throw new RemoteException( e.toString() );
    }
    log.info( "CLIENT > service is registered : " + bean.getClasses()[0] + " hash #" + bean.getHashCode() + " idx=" + idx );
    DebugLog.logRemoteInstance( log, "service(remote)", bean.getRMIService() );
  }

}
