/*
 * Created on Aug 23, 2005
 *
 */
/**
   Copyright (C) 2006 by Alexei Krasnopolski

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License. 
*/
package com.ak.osgi.service.rmi.gate;

import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Vector;

import org.knopflerfish.service.log.LogRef;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.Constants;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.util.tracker.ServiceTrackerCustomizer;

import com.ak.osgi.service.rmi.RmiConstants;
import com.ak.osgi.service.rmi.infobean.RMIServiceBeanImpl;
import com.ak.osgi.service.rmi.interfaces.rmifwcontext.RemoteOSGiFramework;
import com.ak.osgi.service.rmi.remotefw.RMIEntrant;
import com.ak.osgi.service.rmi.remotefw.RMIGate;
import com.ak.osgi.service.rmi.remotefw.RMIServiceBean;
import com.ak.osgi.service.rmi.serviceimpl.rmifwcontext.RemoteOSGiFrameworkImpl;
import com.ak.osgi.service.rmi.utility.DebugLog;

/**
 * The <tt>RemoteServer</tt> class implements RMIGate interface.
 * <p><tt>RemoteServer</tt> object runs on remote framework and serves all <tt>Client</tt> objects
 * these registered by using method <tt>addClient()</tt>.
 * The all services running on this remote framework with property 'rmiservice' are registered
 * on <tt>Client</tt>'s frameworks this moment.
 * <p>New remote services (with 'rmiservice' property) after registering on remote framework
 * will be registered with <tt>Client</tt>'s framework
 * 
 * @version 1.0.0
 * @author <a href="mailto:krasnop@bellsouth.net">Alexei Krasnopolski</a>
 */
public class RemoteServer extends UnicastRemoteObject implements RMIGate
{
  private static final long serialVersionUID = 1L;
  /**
   *  This field is a reference to <tt>BundleActivator</tt> object that was running this object.
   */
  protected com.ak.osgi.service.rmi.Activator activator;
  /**
   *  <tt>Vector</tt> clients keeps remote references to <tt>RMIEntrant</tt> clients of the framework 
   */
  Vector clients = new Vector();
  /**
   *  <tt>BundleContext</tt> of the <tt>RemoteServer</tt> bundle.
   */
  BundleContext bc;
  /**
   *  <tt>ServiceRegistration</tt> object of Remoteserver's service.
   */
  protected ServiceRegistration mySR;
  /**
   *  rmiServiceTracker tracks the all services with 'rmiservice' property.
   */
  protected ServiceTracker rmiServiceTracker;
  /**
   *  customizer is a listener of life-cycle events of remote services on this framework
   */
  protected ServiceTrackerCustomizer customizer;
  /**
   *  RMI registry port
   */
  private static int RMIRegistryPort;
  /**
   *  reference to local RMI registry
   */
  private Registry registry;
  private RemoteOSGiFramework rFWService = null;
  protected LogRef log;

  /**
   * The <tt>RMIServiceTrackerCustomizer</tt> class provides processing 
   * of service life cycle events.
   * <p>If instance of remote service is adding 
   * then this service is registered on all client's frameworks.
   * <p>If remote service is removing from this remote framework then 
   * this service is unregistered from all client's frameworks. 
   * 
   */
  class RMIServiceTrackerCustomizer implements ServiceTrackerCustomizer
  {
    public Object addingService( ServiceReference reference )
    {
      return registerServiceOnRemoteFW( reference, clients.toArray() );
    }

    public void modifiedService( ServiceReference reference, Object service )
    {}

    public void removedService( ServiceReference reference, Object service )
    {        
      unRegisterServiceOnRemoteFW( reference, service, clients.toArray() );
    }
  }
  /**
   * Create instance of <tt>RemoteServer</tt> object
   * 
   * @param context
   *           context of <tt>RemoteServer</tt> bundle 
   * @throws RemoteException
   *           If can not create instance of <tt>UnicastRemoteObject</tt> super class. 
   */
  public RemoteServer( com.ak.osgi.service.rmi.Activator $activator ) throws RemoteException
  {
    super();
    activator = $activator;
    bc = activator.getBundleContext();
    rFWService = new RemoteOSGiFrameworkImpl( bc );
    log = activator.getLogger();
  }

  /**
   * For each connected client this method invokes registerRMIService() method that register 
   * given service on remote client's framework.
   * Before it extracts properties and interfaces from service reference and pass its to
   * remote framework for new registration.
   * 
   * @param reference
   *           <tt>ServiceReference</tt> object for service is registering on client framework. 
   * @param rmiClients
   *           Array of clients connected to this RMI Gate.
   * @return
   *           registered remote service.
   * @throws RemoteException
   *           If any client can not register the remote service. 
   */
  protected Object registerServiceOnRemoteFW( ServiceReference reference, Object [] rmiClients )
  {
    Remote rmiService = (Remote)bc.getService( reference );
    String serviceClassName = rmiService.getClass().getName();
    Object rmiserviceProp = reference.getProperty( RmiConstants.KEY_RMISERVICE_LOC );
    if( rmiClients.length > 0 && rmiserviceProp != null )
    {
      String [] clazzes = (String [])reference.getProperty( Constants.OBJECTCLASS );
      String [] keys = reference.getPropertyKeys();
      String key;
      Dictionary properties = new Hashtable();
      for( int i = 0; i < keys.length; i++ )
      {
        key = keys[i];
        if( key.equalsIgnoreCase( RmiConstants.KEY_RMISERVICE_LOC ) )
        {
          key = RmiConstants.KEY_RMISERVICE_REM;  // new key for old value
        }
        properties.put( key, reference.getProperty( keys[i] ) );
      }
      
      RMIServiceBean bean = new RMIServiceBeanImpl();
      bean.setHashCode( reference.hashCode() );
      bean.setClasses( clazzes );
      bean.setRMIService( rmiService );
      bean.setProperties( properties );
      
      for( int j = 0; j < rmiClients.length; j++ )
      {
        try
        {
          ((RMIEntrant)rmiClients[j]).registerRMIService( bean );
        }
        catch( RemoteException e )
        {
          log.error( "SERVER > Gate exception: while register service " + clazzes[0] 
                                        + " on client's context.", e );
        }
      }
      log.info( "SERVER > register rmi service " + clazzes[0] + " to remote client framework; hash #" + reference.hashCode());
      DebugLog.logRemoteInstance( log, "remote service", rmiService );
    }
    else
      log.debug( "SERVER > can not register service " + serviceClassName + "; prop= " + rmiserviceProp + "; hash #" + reference.hashCode() + " on any remote client fw's" );
    return rmiService;
  }
  
  protected void unRegisterServiceOnRemoteFW( ServiceReference reference, Object service, Object [] rmiClients )
  {
    String serviceClassName = "unknown";
    if( service != null )
      serviceClassName = service.getClass().getName();
    try
    {
      for( int j = 0; j < rmiClients.length; j++ )
      {
        ((RMIEntrant)rmiClients[j]).unregisterRMIService( reference.hashCode() );
        bc.ungetService( reference );
      }
    }
    catch( RemoteException e )
    {
      log.error( "SERVER > can not unregister rmi service from remote framework: " + serviceClassName+ " hash #" + reference.hashCode(), e );        
    }
    log.info( "SERVER > UNregister rmi service from remote framework: " + serviceClassName + " hash #" + reference.hashCode() );
  }

  /**
   * Add client to the <tt>RemoteServer</tt>. If the client already was added the request will be ignored.
   * Then all existing remote services are registered on the client.
   *
   *@param client
   *          <tt>RMIEntrant</tt> object to be added as a client of this framework. 
   *@see com.ak.osgi.service.rmi.remotefw.RMIGate#addClient(com.ak.osgi.service.rmi.remotefw.RMIEntrant)
   */ 
  public void addClient( RMIEntrant client )
  {
    if( !clients.contains( client ) )
    {
      clients.add( client );
      ServiceReference [] srefs = rmiServiceTracker.getServiceReferences();
      log.info( "SERVER > Client added to Gate. # of tracked services =" 
          + ((srefs==null)? -1 : srefs.length) );
      DebugLog.logRemoteInstance( log, "client(remote)", client );
      for( int i = 0; i < srefs.length; i++ )
      {
        registerServiceOnRemoteFW( srefs[i], new RMIEntrant[] {client} );
      }
    }
  }
  
  /**
   * Remove client from this <tt>RemoteServer</tt>.
   * 
   *@param client
   *          Reference to removing client.
   *@see com.ak.osgi.service.rmi.remotefw.RMIGate#removeClient(com.ak.osgi.service.rmi.remotefw.RMIEntrant)
   */ 
  public void removeClient( RMIEntrant client )
  {
    ServiceReference [] srefs = rmiServiceTracker.getServiceReferences();
    for( int i = 0; i < srefs.length; i++ )
    {
      unRegisterServiceOnRemoteFW( srefs[i], null, new RMIEntrant[] {client} );
    }
    clients.remove( client );
    log.info( "SERVER > Remote Client has removed from the Gate." );
  }
  
  /**
   * Open the <tt>RemoteServer</tt> object:
   *  <br>1. Create and open service tracker for any remote service on this framework,
   *  <br>2. Register itself as a remote service. It causes register <tt>RMIGate</tt> 
   *  service on all client remote frameworks
   * @throws BundleException
   *           If <tt>ServiceTracker</tt> object can not be created or opened. 
   *       <p>IllegalArgumentException
   *       <p>SecurityException
   *       <p>IllegalstateException
   *           are throwed from registerService(...) method.
   *             
   */
  public void open() throws BundleException
  {
    customizer = new RMIServiceTrackerCustomizer();
// create and run tracker for all remote services (with property 'rmiservice')    
    try
    {
      String rmiServiceFilter = "(" + RmiConstants.KEY_RMISERVICE_LOC + "=*)";
      rmiServiceTracker = new ServiceTracker( bc, bc.createFilter( rmiServiceFilter ), customizer );
      rmiServiceTracker.open();
    }
    catch( InvalidSyntaxException e )
    {
      throw new BundleException( "SERVER > Failed to set up rmiService tracker: " + e );
    }
// register itself as a remote service on local FW (and on all remote FW's too)
    Dictionary props = new Hashtable();
    props.put( RmiConstants.KEY_RMISERVICE_LOC, "RMI Gate" );
    String [] clazzes = new String[] { RMIGate.class.getName() };
    mySR = bc.registerService( clazzes, this, props );
    DebugLog.logRemoteInstance( log, "remoteServer(local)", this );
  }

  /**
   * Close the <tt>RemoteServer</tt> object. Unregester own service and close service tracker.
   * Unbind <tt>RemoteServer</tt> object from RMI registry.
   *  
   * @throws RemoteException, NotBoundException
   *            If can not unbind 'RMIGate' object from RMI registry
   */
  public void close() throws RemoteException, NotBoundException
  {
    try
    {
      mySR.unregister();
    }
    catch( IllegalStateException e )
    {}
    rmiServiceTracker.close();
    registry.unbind( "RMIGate" );
  }

  /**
   * Link <tt>RemoteServer</tt> object to RMI registry. If can not access RMI registry then
   * try to run it.
   * 
   * @param $RMIRegistryPort
   *           port number of local RMI registry
   * @return
   *           <tt>true</tt> if successfuly completed
   */          
  public boolean setup( int $RMIRegistryPort )
  {
    RemoteServer.RMIRegistryPort = $RMIRegistryPort;
    boolean cc = false;
    try
    {
      registry = LocateRegistry.getRegistry( RMIRegistryPort );
    }
    catch( RemoteException e )
    {
      log.error( "SERVER > RMI exception: can not find local rmi registry.", e );
      return cc;
    }
    
  //  Bind the remote object's stub to the RMI registry. Trying to bind two times:
  //  if after first trying catch exception then start RMI registry
  //  and try to bind second time.  
    for( int trying = 0; trying < 2; trying++ )
    {
      try
      {
        registry.bind( "RMIGate", this );
        cc = true;
        break;
      }
      catch( RemoteException e )
      {
        String msg = "SERVER > RMI exception: can not access rmi registry while bind.";
        if( (trying >= 1 || (registry = startRMIRegistry()) == null) )
        {
          log.error( msg, e );
          break; 
        }
//        log.warn( msg );
      }
      catch( AlreadyBoundException e )
      {
        log.warn( "SERVER > RMI exception: 'RMIGate' object already is binded to rmi registry.\n" );
        try
        {
          registry.rebind( "RMIGate", this );
          cc = true;
          break;
        }
        catch( RemoteException e2 )
        {
          log.error( "SERVER > RMI exception: can not access rmi registry while rebind.", e2 );
        }
      }
    }
    return cc;
  }

  /**
   * Start RMI registry on local host.
   * 
   * @return
   *       created RMI registry.
   */
  private Registry startRMIRegistry()
  {
    Registry rgs = null;
    try
    {
      rgs = LocateRegistry.createRegistry( RMIRegistryPort );
      log.info( "SERVER > RMI registry is running now! java.rmi.server.codebase=" + System.getProperty( "java.rmi.server.codebase" ) );
      Thread.sleep( 1000 );
    }
    catch( RemoteException e )
    {
      log.error( "SERVER > RMI exception: can not run local rmi registry.", e );
    }
    catch( InterruptedException e )
    {}
    return rgs;
  }

  /**
   *@return
   *@throws RemoteException
   *@see com.ak.osgi.service.rmi.remotefw.RMIGate#getRemoteContext()
   */ 
  
  public RemoteOSGiFramework getRemoteContext() throws RemoteException
  {
    return rFWService;
  }

}
