/*
 * Created on Aug 22, 2005
 *
 */
/**
   Copyright (C) 2006 by Alexei Krasnopolski

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License. 
*/
package com.ak.osgi.service.rmi.http;

import java.net.URL;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.knopflerfish.service.log.LogRef;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.Constants;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;
import org.osgi.service.http.NamespaceException;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.util.tracker.ServiceTrackerCustomizer;

/**
 * The <tt>StubDownloader</tt> class create http context that acts as a
 * root of code base for RMI classloader. The context register with running
 * HTTP server (need to be running on both client and server frameworks).
 * After http context successfuly register a "java.rmi.server.codebase"
 * property is added to <tt>System</tt> object. 
 * 
 * @version 1.0.0
 * @author <a href="mailto:krasnop@bellsouth.net">Alexei Krasnopolski</a>
 */
public class StubDownloader
{
  /**
   *  This bundle context
   */
  BundleContext bc;
  /**
   *  Http context for rmi codebase
   */
  HttpContext context;
  /**
   *  HTTP server port
   */
  Integer serverPort;
  /**
   *  serverID is a framework id of running HTTP service
   */
  Long serverID;
  /**
   *  Service tracker of HTTP services
   */
  ServiceTracker httpTracker;
  /**
   *  customizer for <tt>ServiceTracker</tt>
   */
  ServiceTrackerCustomizer customizer;
  /**
   *  The field helps to control of the object after STOPPING event   
   */
  boolean isStoppingMode = false;
  /**
   *  Path to RMI codebase root directory (relative to bundle jar root directory)
   */
  static final String RES_DIR = ""; 
  /**
   *  http context alias of RMI codebase
   */
  static final String RES_ALIAS = "/codebase";
  int defaultHttpPort;
  static LogRef log;
  /**
   * Create <tt>StubDownloader</tt> instance.
   * Create http context object.
   * Create HTTP service tracker with custom listener. 
   * 
   * @param $bc
   *          Bundle context
   */
  public StubDownloader( BundleContext $bc )
  {
    bc = $bc;
    log = new LogRef( bc );
    serverID = null;
// take the property for http port from properties file only for Oscar framework 
// for others take it from really running HTTP service 
    try
    {
      defaultHttpPort = Integer.parseInt( bc.getProperty( "org.osgi.service.http.port" ) );
    }
    catch (Exception e)
    {
      defaultHttpPort = 8080;
    }
    context = new HttpContext()
    {
      public URL getResource( String name )
      {
        URL url = getClass().getResource( name );
        return url;
      }

      public String getMimeType( String name )
      {
        return null;
      }
      
      public boolean handleSecurity( HttpServletRequest request,
                                      HttpServletResponse response ) 
      {
        return true;
      }
      
    };

    customizer = new ServiceTrackerCustomizer()
    {
      /**
       * Adding HTTP service to tracker.
       * The method retrieves port number and service id properties from
       * new HTTP service. Then register own http context on
       * this http service and set System property 'java.rmi.server.codebase'
       * that is needed for RMI Class Loader  
       * 
       *@param reference
       *         Service reference
       *@return
       *         added service
       *@see org.osgi.util.tracker.ServiceTrackerCustomizer#addingService(org.osgi.framework.ServiceReference)
       */ 
      public Object addingService( ServiceReference reference )
      {
        HttpService httpServer = (HttpService)bc.getService( reference );
        if( serverID == null )  // context is registered for only one HTTP server.
        try
        {
// HTTP service from Knopflerfish has 'port.http' property key
          serverPort = (Integer)reference.getProperty( "port.http" ); 
// HTTP service from Equinox has 'http.port' property key
          if( serverPort == null ) 
            serverPort = (Integer)reference.getProperty( "http.port" ); 
// HTTP service from Oscar has no properties at all !!!
          if( serverPort == null )
            serverPort = new Integer( defaultHttpPort );
          serverID = (Long)reference.getProperty( Constants.SERVICE_ID );
          httpServer.registerResources( RES_ALIAS, RES_DIR, context );
          System.setProperty( "java.rmi.server.codebase", 
            "http://localhost:" + serverPort + RES_ALIAS +"/");
          log.info( "StubDownloader : register codebase context to HTTP server : " 
              + System.getProperty( "java.rmi.server.codebase" ) );
        }
        catch( SecurityException e )
        {
          log.error( "StubDownloader : can not set system property.", e );
        }
        catch( NamespaceException e )
        {
          log.error( "StubDownloader : can not register http context.", e );          
        }
        return httpServer;
      }

      public void modifiedService( ServiceReference reference, Object service )
      {}

      /**
       * Removing HTTP service from the tracker.
       * If removed service id equals to HTTP service id, with that this context was registered
       * unregister codebase context and clears system property. 
       * 
       *@param reference
       *@param service
       *@see org.osgi.util.tracker.ServiceTrackerCustomizer#removedService(org.osgi.framework.ServiceReference, java.lang.Object)
       */ 
      public void removedService( ServiceReference reference, Object service )
      {        
        Long id = (Long)reference.getProperty( Constants.SERVICE_ID );
        if( serverID != null && id.intValue() == serverID.intValue() )
        {
          HttpService httpServer = (HttpService)service;
          try
          {
            httpServer.unregister( RES_ALIAS );
            System.setProperty( "java.rmi.server.codebase", "" );
            log.info( "StubDownloader : UNregister codebase context from HTTP server" );
            serverID = null;
            if( !isStoppingMode )
              bc.getBundle().stop();
          }
          catch( BundleException e )
          {
            log.error( "StubDownloader : error while stopping the bundle", e );
          }
          catch( IllegalArgumentException iae )
          {
            log.error( "StubDownloader : there is no registration for the " + RES_ALIAS, iae );
          }
        }  
      }
    };
    
    try
    {
      String httpFilter = "(objectclass=org.osgi.service.http.HttpService)";
      httpTracker = new ServiceTracker( bc, bc.createFilter( httpFilter ), customizer );
    }
    catch( InvalidSyntaxException e )
    {
      throw new RuntimeException( "StubDownloader : failed to setup http tracker: " + e );
    }
    
  }
  
  /**
   * Open http service tracker and check if exists at least one.
   */
  public void register()
  {
    httpTracker.open();
    ServiceReference [] allServers = httpTracker.getServiceReferences();
    
    if( allServers == null )
    {
      throw new RuntimeException( "StubDownloader : can not find any http server." );
    }
  }

  /**
   * Close http service tracker and prepare the objects for GC
   */
  public void unregister()
  {
    isStoppingMode = true;
    httpTracker.close();
    httpTracker = null;
    customizer = null;
  }
  
}

//String [] keys = reference.getPropertyKeys();
//System.out.println( "************ properties **********" );
//for( int i = 0; i < keys.length; i++ )
//{
//String key = keys[i];
//Object value = reference.getProperty( key );
//if( value instanceof String [] )
//{
//  String [] values = (String [])value;
//  System.out.println( "entry["+ i + "] : " + key + " = " + value );
//  for( int j = 0; j < values.length; j++ )
//    System.out.println( "        [" + j + "] = " + values[j] );
//}
//else
//  System.out.println( "entry["+ i + "] : " + key + " = " + value );
//}
